#!/usr/bin/perl -w
#
# Patchwork - automated patch tracking system
# Copyright (C) 2005 Jeremy Kerr <jk@ozlabs.org>
# - Modified by David Howells <dhowells@redhat.com>
#
# This file is part of the Patchwork package.
#
# Patchwork is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Patchwork is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Patchwork; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use strict;
use lib '../lib';

use PatchWork::PatchParser;
use PatchWork::SiteConfig;
use PatchWork::PatchDatabase;
use Getopt::Long;

my $usecurtime = undef;
my $config = undef;
my $ignorelistid = undef;

exit 1 unless GetOptions('curtime' => \$usecurtime,
			 'config=s' => \$config,
			 'ignorelistid' => \$ignorelistid);

$ENV{PATCHWORK_CONFIG} = $config if $config;

my $siteconfig = new PatchWork::SiteConfig() or
    die "No such config $config";

my $options = {
	UseCurTime => $usecurtime,
	IgnoreListID => $ignorelistid
    };

#
# parse the message
#
my $patchparser = new PatchWork::PatchParser($siteconfig);

if ($#ARGV == 0) {
	$patchparser->parse_MIME_file(shift @ARGV, $options);
}
else {
	$patchparser->parse_MIME_stream(\*STDIN, $options);
}

my $submission = $patchparser->submission();
my @patchfiles = $patchparser->patchfiles();
my @comments = $patchparser->comments();

exit unless (@patchfiles || @comments);

#
# okay... we've got something to add, so we need the database
#
my $db = new PatchWork::PatchDatabase($siteconfig);

#
# get a personnel descriptor for this email's submitter and attach to all
# interested descriptors
#
my $submitter = $db->lookup_person($patchparser->submitter());

$submission->submitter($submitter) if $submission;
$_->submitter($submitter) foreach (@comments);

if ($submission) {
	# map the project list to database row IDs
	my @projects = $patchparser->projectnames();
	@projects = map { $db->lookup_project($_) } @projects;

	unless (@projects) {
		my $nullproj = new PatchWork::Project();
		$nullproj->id(0);
		push @projects, $nullproj;
	}

	# generate a set of patch references
	$db->begin();

	$db->insert_submission($submission);

	my $bug = $patchparser->bug();
	$db->insert_bug($bug) if $bug;

	$db->insert_patchfile($_) foreach (@patchfiles);

	foreach my $project (@projects) {
		foreach my $patchfile (@patchfiles) {
			my $patch = new PatchWork::Patch();
			$patch->patchfile($patchfile);
			$patch->project($project) if $project->id();
			$patch->bug($bug);
			$db->insert_patch($patch);
		}
	}

	$db->insert_comment($_) foreach (@comments);

	$db->commit();
}
elsif (@comments) {
	$db->begin();

	# find the submitted patch to which this corresponds if there is one
	$submission = $db->lookup_submission($patchparser->refs());
	exit unless $submission;

	$db->mark_patches_under_review($submission);

	$_->submission($submission) foreach (@comments);

	$db->insert_comment($_) foreach (@comments);

	$db->commit();
}
