# Patchwork - automated patch tracking system
# Copyright (C) 2005 Jeremy Kerr <jk@ozlabs.org>
# - Modified by David Howells <dhowells@redhat.com>
#
# This file is part of the Patchwork package.
#
# Patchwork is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Patchwork is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Patchwork; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package PatchWork::PatchDatabase;

use strict;
use PatchWork::Person;
use PatchWork::User;
use PatchWork::Submission;
use PatchWork::PatchFile;
use PatchWork::Patch;
use PatchWork::Comment;
use PatchWork::Project;
use PatchWork::Bug;
use PatchWork::SiteConfig;
use PatchWork::Tree;
use DBI;
use Carp;
use Digest::MD5 qw(md5_hex);
use Text::ParseWords;

# internal variables
#  config		PatchWork::SiteConfig
#  dbh			our database handle
#  cache_projects	cache of project records by pretty name
#  cache_projects_id	cache of project records by row ID
#  cache_people		cache of people by email addr
#  cache_people_id	cache of people by row ID
#  trees		cache of trees


my %order_tbl = (
	     default	=> [ 'states.ordering', 'submissions.date DESC' ],
	     subject	=> [ 'submissions.subject', 'patches.project' ],
	     patch	=> [ 'submissions.subject', 'patchfiles.patch' ],
	     date	=> [ 'submissions.date DESC' ],
	     submitter	=> [ 'people.name', 'people.email' ],
	     project	=> [ 'projects.name', 'projects.version' ],
	     ref	=> [ 'bugreports.type', 'bugreports.ref' ],
	     state	=> [ 'states.ordering' ]
	     );

sub new($$)
{
	my ($cls, $config) = @_;
	my $obj = {};
	bless($obj, $cls);
	$obj->{config} = $config;
	$obj->{cache_projects} = {};
	$obj->{cache_projects_id} = {};
	$obj->{cache_people} = {};
	$obj->{cache_people_id} = {};

	my ($user, $pass) = $obj->{config}->db_auth();
	my $dbh = DBI->connect($obj->{config}->dsn(), $user, $pass,
			       { RaiseError => 0,
				 AutoCommit => 0
			     } )
	    or die("Couldn't connect to database (" .
		   $obj->{config}->dsn() .
		   "): " .
		   $DBI::errstr);

	$obj->{dbh} = $dbh;

	$obj->_load_projects();
	$obj->_load_personnel();

	return $obj;
}

sub DESTROY($)
{
	my ($obj) = @_;
	if ($obj->{dbh}) {
		$obj->{dbh}->disconnect();
	}
}

sub begin($)
{
	my ($obj) = @_;
	#$obj->{dbh}->begin_work();
}

sub commit($)
{
	my ($obj) = @_;
	$obj->{dbh}->commit();
}

sub rollback($)
{
	my ($obj) = @_;
	$obj->{dbh}->rollback();
}

sub _qi(@)
{
	my ($obj) = shift;
	return $obj->{dbh}->quote_identifier(@_);
}

###############################################################################
#
# A convenience method for calling a single prepare->execute pair
#
###############################################################################
sub _query($$@)
{
	my ($obj, $sql, @values) = @_;
	my $sth = $obj->{dbh}->prepare($sql)
		or croak "Failed prepare ($sql): ".$obj->{dbh}->errstr();
	$sth->execute(@values)
		or croak "Failed execute ($sql): ".$obj->{dbh}->errstr();
	return $sth;
}

###############################################################################
#
# the Postgres DBI driver doesn't support last_insert_id, so we have to
# wrap the calls with our own function that will provide an alternate
# when we're using the DBD::Pg driver :(
#
# Also, mysql appears to only support $sth->{mysql_insertid}
#
###############################################################################
sub _insert_id($$$$)
{
	my ($obj, $sth, $table, $col) = @_;

	my $id;

	if ($obj->{dbh}->{Driver}->{Name} eq 'Pg') {
		$sth = $obj->_query("SELECT $col FROM $table WHERE oid=?",
				$sth->{pg_oid_status});
		if (my $row = $sth->fetchrow_hashref()) {
			$id = $row->{id};
		} else {
			croak "Couldn't retrieve last inserted id";
		}
	} elsif ($obj->{dbh}->{Driver}->{Name} eq 'mysql') {
		return $sth->{mysql_insertid};

	} else {
		$id = $obj->{dbh}->last_insert_id(undef, undef, $table, $col)
			or croak "Couldn't retrieve last inserted id";
	}

	return $id;
}

###############################################################################
#
# Format a date appropriately
# - MySQL doesn't support 'epoch' in the extract function
#
###############################################################################
sub _sql_epoch($$)
{
	my ($obj, $col) = @_;

	my $fmt = 'EXTRACT(epoch FROM %s)';
	if ($obj->{dbh}->{Driver}->{Name} eq 'mysql') {
		$fmt = 'UNIX_TIMESTAMP(%s)';
	}
	return sprintf($fmt, $col);
}

###############################################################################
#
# Get the state list
#
###############################################################################
sub get_states($)
{
	my ($obj) = @_;
	my $sth = $obj->_query('SELECT id, name FROM states');
	my %states;
	while (my $row = $sth->fetchrow_hashref()) {
		$states{$row->{id}} = $row->{name};
	}
	return %states;
}

###############################################################################
#
# Load the project list
#
###############################################################################
sub _load_projects($)
{
	my ($obj) = @_;

	my $sth = $obj->_query('SELECT id, name, version, alias FROM projects');
	my $rows = $sth->fetchall_arrayref();

	foreach my $row (@$rows) {
		my $project = new PatchWork::Project();
		$project->id($row->[0]);
		$project->set($row->[1], $row->[2], $row->[3]);
		$obj->{cache_projects}->{$project->pretty()} = $project;
		$obj->{cache_projects_id}->{$project->id()} = $project;
	}

	$obj->rollback();
}

###############################################################################
#
# Look up a project descriptor by pretty name
#
###############################################################################
sub lookup_project($$)
{
	my ($obj, $pretty) = @_;

	return $obj->{cache_projects}->{$pretty} ||
	    die "Unknown project $pretty\n";
}

###############################################################################
#
# Look up a project descriptor by row ID
#
###############################################################################
sub lookup_project_id($$)
{
	my ($obj, $id) = @_;

	return $obj->{cache_projects_id}->{$id};
}

###############################################################################
#
# Get a list of known projects
#
###############################################################################
sub get_project_list($)
{
	my ($obj) = @_;

	return %{$obj->{cache_projects}};
}

###############################################################################
#
# Load the personnel list
#
###############################################################################
sub _load_personnel($)
{
	my ($obj) = @_;

	my $sth = $obj->_query('SELECT id, email, name FROM people');
	my $rows = $sth->fetchall_arrayref();

	foreach my $row (@$rows) {
		my $person = new PatchWork::Person($row->[1], $row->[2], $row->[0]);
		$obj->{cache_people}->{$person->email()} = $person;
		$obj->{cache_people_id}->{$person->id()} = $person;
	}

	$obj->rollback();
}

###############################################################################
#
# Look up a personnel descriptor in the cache by ID
#
###############################################################################
sub lookup_person_id($$)
{
	my ($obj, $id) = @_;

	return $obj->{cache_people_id}->{$id};
}

###############################################################################
#
# Look up a personnel descriptor in the cache, creating one if it doesn't exist
# - this ends any outstanding transaction
#
###############################################################################
sub lookup_person($$$)
{
	my ($obj, $email, $name) = @_;

	# return a pre-existing record if we can
	return $obj->{cache_people}->{$email} if $obj->{cache_people}->{$email};

	# doesn't exist; so we'll attempt to add it to the database
	# - because we can't use table locking (there's no way that will work
	#   on both PostgreSQL and MySQL), try to insert a row. If that fails,
	#   assume we've hit a race with the same person (there's a unique
	#   constraint on the email column), and lookup the email again.
	my ($sql, @values);

	if ($name) {
		$sql = 'INSERT INTO people (email, name) VALUES (?,?)';
		@values = ($email, $name);
	} else {
		$sql = 'INSERT INTO people (email) VALUES (?)';
		@values = ($email);
	}

	my $id;
	my $sth = $obj->{dbh}->prepare($sql)
	    or croak "Failed prepare ($sql): " . $obj->{dbh}->errstr();

	if ($sth->execute(@values)) {
		$id = $obj->_insert_id($sth, 'people', 'id');
		$obj->{dbh}->commit();
	}
	else {
		$obj->{dbh}->rollback();

		$sth = $obj->_query('SELECT id FROM people WHERE email=?',
				    $email);

		if (my $row = $sth->fetchrow_hashref()) {
			$id = $row->{id};
			$obj->{dbh}->rollback();
		}
		else {
			# ruh-roh.
			die "Secondary SELECT failed finding person";
		}
	}

	# create a person descriptor and cache it
	my $person = new PatchWork::Person($email, $name, $id);
	$obj->{cache_people}->{$email} = $person;
	$obj->{cache_people_id}->{$id} = $person;
	return $person;
}

###############################################################################
#
# Get a list of known people
#
###############################################################################
sub get_people_list($)
{
	my ($obj) = @_;
	return %{$obj->{cache_people}};
}

###############################################################################
#
# Add a patch submission to the submissions table
# - caller must commit
#
###############################################################################
sub insert_submission($$)
{
	my ($obj, $submission) = @_;

	my $sth = $obj->_query('INSERT INTO submissions' .
			       ' (msgid, subject, date, submitter, npatches) '.
			       'VALUES (' .
			       "?, " .
			       "?, " .
			       "'" . $obj->_date_format($submission->date()) . "', " .
			       "" . $submission->submitter()->id() . ", " .
			       "" . $submission->npatches() . "" .
			       ")",
			       $submission->msgid(),
			       $submission->subject());

	return $submission->id($obj->_insert_id($sth, 'submissions', 'id'));
}

###############################################################################
#
# Look up a patch submission by message ID
#
###############################################################################
sub lookup_submission($@)
{
	my ($obj, $forupdate, @refs) = @_;

	return undef unless @refs;

	my $sql = 'SELECT id FROM submissions WHERE ' .
		join(' OR ', map { "msgid=?" } @refs);

	my $sth = $obj->_query($sql, @refs);
	my $row = $sth->fetchrow_arrayref() or return undef;

	my $submission = new PatchWork::Submission();
	$submission->id($row->[0]);

	return $submission;
}

###############################################################################
#
# Delete a submission
# - caller must commit
#
###############################################################################
sub delete_submission($$)
{
	my ($obj, $submission) = @_;

	return undef unless $submission;
	$obj->_query('DELETE FROM submissions WHERE id=?', $submission->id());
	return 1;
}

###############################################################################
#
# Add a patch file to the patchfiles table
# - caller must commit
#
###############################################################################
sub insert_patchfile($$)
{
	my ($obj, $file) = @_;

	my $sth = $obj->_query('INSERT INTO patchfiles' .
			       ' (submission, patch, filename, content) '.
			       'VALUES (' .
			       "" . $file->submission()->id() . ", " .
			       "" . $file->patchix() . ", " .
			       "?, " .
			       "?" .
			       ")",
			       $file->filename(),
			       $file->content());

	return $file->id($obj->_insert_id($sth, 'patchfiles', 'id'));
}

###############################################################################
#
# Count the number of patchfiles remaining for a submission
#
###############################################################################
sub count_patchfiles($$)
{
	my ($obj, $submission) = @_;
	my $sth = $obj->_query('SELECT id FROM patchfiles WHERE submission=?',
			       $submission->id());

	my $count = 0;
	$count++ while ($sth->fetchrow_arrayref());

	return $count;
}

###############################################################################
#
# Delete a patch file
# - caller must commit
#
###############################################################################
sub delete_patchfile($$)
{
	my ($obj, $patchfile) = @_;

	return undef unless $patchfile;
	$obj->_query('DELETE FROM patchfiles WHERE id=?', $patchfile->id());
	return 1;
}

###############################################################################
#
# Add a patch to the patches table
# - caller must commit
#
###############################################################################
sub insert_patch($$)
{
	my ($obj, $patch) = @_;

	my $sth = $obj->_query('INSERT INTO patches' .
			       ' (state, project, bug, content) '.
			       'VALUES (' .
			       "" . $patch->state() . ", " .
			       "" . $obj->_project_format($patch->project()) . ", " .
			       "" . $obj->_bug_format($patch->bug()) . ", " .
			       "" . $patch->patchfile()->id() . "" .
			       ")");

	return $patch->id($obj->_insert_id($sth, 'patches', 'id'));
}

###############################################################################
#
# Change a patch's state
# - caller must commit
#
###############################################################################
sub update_patch($$)
{
	my ($obj, $patch) = @_;

	my $project = $patch->project();
	my $bug = $patch->bug();

	$obj->_query('UPDATE patches'.
		     ' SET '.
		     '	state=' . $patch->state() . ',' .
		     '	archived=' . ($patch->archived() ? 'true' : 'false') . ',' .
		     '  project=' . ($project ? $project->id() : 'NULL') . ',' .
		     '  bug=' . ($bug ? $bug->id() : 'NULL') .
		     ' WHERE id=' . $patch->id()
		     );

	return 1;
}

###############################################################################
#
# Count the number of patches remaining for a patch file
#
###############################################################################
sub count_patches($$)
{
	my ($obj, $patchfile) = @_;
	my $sth = $obj->_query('SELECT id FROM patches WHERE content=?',
			       $patchfile->id());

	my $count = 0;
	$count++ while ($sth->fetchrow_arrayref());

	return $count;
}

###############################################################################
#
# Delete a patch
# - caller must commit
#
###############################################################################
sub delete_patch($$)
{
	my ($obj, $patch) = @_;

	return undef unless $patch;
	$obj->_query('DELETE FROM patches WHERE id=?', $patch->id());
	return 1;
}

###############################################################################
#
# Find the patches that match a particular submission and mark them under as
# review
# - caller must commit
#
###############################################################################
sub mark_patches_under_review($$)
{
	my ($obj, $submission) = @_;

	my $sth = $obj->_query('SELECT patches.id, patches.state' .
			       ' FROM patches' .
			       ' INNER JOIN patchfiles ON patches.content = patchfiles.id' .
			       ' WHERE submission=?' .
			       ' FOR UPDATE',
			       $submission->id());

	while (my $row = $sth->fetchrow_arrayref()) {
		$obj->_query('UPDATE patches SET state=2 WHERE id=' . $row->[0])
		    if $row->[1] == 1;
	}
}

###############################################################################
#
# Add a comment to the comment table
# - caller must commit
#
###############################################################################
sub insert_comment($$)
{
	my ($obj, $comment) = @_;

	my $date = $obj->_date_format($comment->date());

	my $sth = $obj->_query('INSERT INTO comments' .
			       ' (submission, msgid, submitter, date, content)'.
			       'VALUES (' .
			       $comment->submission()->id() . ', ' .
			       "?, " .
			       "" . $comment->submitter()->id() . ", " .
			       "'" . $date . "', " .
			       "?" .
			       ")",
			       $comment->msgid(),
			       $comment->content());

	return $comment->id($obj->_insert_id($sth, 'comments', 'id'));
}

###############################################################################
#
# Add a bug reference into the database
# - caller must commit
#
###############################################################################
sub insert_bug($$)
{
	my ($obj, $bug) = @_;

	# see if there's a match already in the bug report list
	my $sel = $obj->_query('SELECT id FROM bugreports WHERE type=? AND ref=?',
			       $bug->type(), $bug->ref());

	if (my $row = $sel->fetchrow_hashref()) {
		return $bug->id($row->{id});
	}
	else {
		# there wasn't; add a new bug report reference
		$row = undef;
		$sel = undef;

		my $ins = $obj->_query("INSERT INTO bugreports (type, ref) VALUES(?, ?)",
				       $bug->type(), $bug->ref());

		return $bug->id($obj->_insert_id($ins, 'bugreports', 'id'));
	}
}

###############################################################################
#
# generate a set of restriction clauses for an SQL SELECT statement
#
###############################################################################
sub _sql_restrictions($$)
{
	my ($obj, $restrictions) = @_;
	my $sql = "";
	my @wheres = ();
	my @orders = ();
	my @bindings = ();

	# impose restrictions
	if (my $patchid = $$restrictions{patchid}) {
		push @wheres, 'patches.id = ' . $patchid;
	}
	else {
		if (my $archived = $$restrictions{archived}) {
			push @wheres, 'patches.archived = \'true\'';
		}
		else {
			push @wheres, 'patches.archived = \'false\'';
		}

		if (my $project = $$restrictions{project}) {
			push @wheres, 'patches.project = ' . $project;
		}

		if (my $submitter = $$restrictions{submitter}) {
			push @wheres, 'submissions.submitter = ' . $submitter;
		}

		if (my $filter = $$restrictions{filter}) {
			my @words = shellwords($filter);
			@words = ($filter) unless @words;
			push @wheres, map { 'submissions.subject ILIKE ?' } @words;
			push @bindings, map { "%".$_."%" } @words;
		}

		if (my $bugfilter = $$restrictions{bugfilter}) {
			if ($bugfilter eq 'set') {
				push @wheres, 'patches.bug IS NOT NULL';
			}
			elsif ($bugfilter eq 'unset') {
				push @wheres, 'patches.bug IS NULL';
			}
			else {
				my @terms = split /[\s,]/, $bugfilter;
				my @types = grep /^\D/, @terms;
				my @refs = grep /^\d/, @terms;

				push @wheres, '(' .
				    join(' OR ', map { "bugreports.type ILIKE ?" } @types) .
				    ')';
				push @bindings, map { "%".$_."%" } @types;

				push @wheres, '(' .
				    join(' OR ', map { 'bugreports.ref LIKE ?' } @refs) .
				    ')';
				push @bindings, @refs;
			}
		}

		# impose ordering constraints
		if (my $name = $$restrictions{order}) {
			push @orders, @{$order_tbl{$name}};
		}

		push @orders, @{$order_tbl{'default'}};
		$sql .= ' ORDER BY ' . join(', ', @orders) . ' ';

		# control quantity of results
		if (my $limit = $$restrictions{limit}) {
			$sql .= " LIMIT $limit ";
		}

		if (my $offset = $$restrictions{offset}) {
			$sql .= " OFFSET " . ($offset - 1) . " ";
		}
	}

	$sql = ' WHERE ' . join(' AND ', @wheres) . ' ' . $sql;

	return ($sql, @bindings);
}

###############################################################################
#
# Construct a set of patch objects from a DB row
#
###############################################################################
sub _patch_from_row($$)
{
	my ($obj, $row) = @_;

	my $submission = new PatchWork::Submission();
	$submission->id($row->{submission});
	$submission->msgid($row->{msgid});
	$submission->subject($row->{subject});
	$submission->date($row->{date});
	$submission->submitter($obj->{cache_people_id}->{$row->{submitter}});
	$submission->npatches($row->{npatches});

	my $file = new PatchWork::PatchFile();
	$file->submission($submission);
	$file->patchix($row->{patch});
	$file->id($row->{file});
	$file->filename($row->{filename});
	$file->content($row->{content});

	my $patch = new PatchWork::Patch();
	$patch->patchfile($file);
	$patch->id($row->{id});
	$patch->state($row->{state});
	$patch->archived($row->{archived} && $row->{archived} ne 'f');
	$patch->project($obj->{cache_projects_id}->{$row->{project}}) if $row->{project};

	if ($row->{bug_type} && $row->{bug_ref}) {
		my $bug = new PatchWork::Bug();
		$bug->set($row->{bug_type}, $row->{bug_ref});
		$patch->bug($bug);
	}

	return $patch;
}

###############################################################################
#
# select a set of patches
#
###############################################################################
sub get_patches(@)
{
	my ($obj, $restrictions) = @_;

	my $sql =
	    'SELECT patches.id,'						.
	    '	submissions.id				AS submission,'		.
	    '	submissions.msgid,'						.
	    '	submissions.subject,'						.
	    $obj->_sql_epoch('submissions.date') . '	AS date,'		.
	    '	submissions.submitter			AS submitter,'		.
	    '	submissions.npatches,'						.
	    '	patchfiles.id				AS file,'		.
	    '	patchfiles.patch,'						.
	    '	patchfiles.filename,'						.
	    '	patchfiles.content,'						.
	    '	patches.state,'							.
	    '	patches.archived, '						.
	    '	patches.project, '						.
	    '   bugreports.type				AS bug_type,'		.
	    '   bugreports.ref				AS bug_ref'		.
	    ' FROM submissions'							.
	    ' INNER JOIN patchfiles ON patchfiles.submission = submissions.id'	.
	    ' INNER JOIN patches ON patches.content = patchfiles.id'		.
	    ' INNER JOIN states ON states.id = patches.state'			.
	    ' INNER JOIN people ON people.id = submissions.submitter'		.
	    ' LEFT JOIN projects ON projects.id = patches.project'		.
	    ' LEFT JOIN bugreports ON bugreports.id = patches.bug'		;

	my ($sqladd, @bindings) = $obj->_sql_restrictions($restrictions);
	$sql .= $sqladd;

	#print "<p>", $sql, "</p>\n";

	my $sth = $obj->_query($sql, @bindings);
	my @patches;

	while (my $row = $sth->fetchrow_hashref()) {
		push(@patches, $obj->_patch_from_row($row));
	}

	return @patches;
}

###############################################################################
#
# Get a specific patch
#
###############################################################################
sub get_patch(@)
{
	my ($obj, $patchid) = @_;

	my %restrictions = ( patchid => $patchid );

	my @patches = $obj->get_patches(\%restrictions);

	return $patches[0];
}

###############################################################################
#
# Get the comments associated with a patch submission
#
###############################################################################
sub get_comments($$)
{
	my ($obj, $submission) = @_;

	my $sql =
	    'SELECT '								.
	    '	id,'								.
	    '	msgid,'								.
	    '	content,'							.
	    '   ' . $obj->_sql_epoch('comments.date') . '	AS date,'	.
	    '	submitter					AS submitter'	.
	    ' FROM comments'							.
	    ' WHERE comments.submission=?'					.
	    ' ORDER BY date'							;

	my $sth = $obj->_query($sql, $submission->id());

	my @comments;

	while (my $row = $sth->fetchrow_hashref()) {
		my $comment = new PatchWork::Comment();
		$comment->id($row->{id});
		$comment->msgid($row->{msgid});
		$comment->content($row->{content});
		$comment->date($row->{date});
		$comment->submitter($obj->lookup_person_id($row->{submitter}));
		push(@comments, $comment);
	}

	return @comments;
}

###############################################################################
#
# Look up a user by some criterion
#
###############################################################################
sub _lookup_user($$@)
{
	my ($obj, $cond, @values) = @_;
	
	my $sql =
	    'SELECT'						.
	    '	users.id,'					.
	    '	users.username,'				.
	    '	people.name,'					.
	    '	people.email'					.
	    ' FROM users'					.
	    ' INNER JOIN people ON people.id = users.id'	.
	    ' WHERE ' . $cond;
	
	my $sth = $obj->_query($sql, @values);
	my $row;
	return undef unless $row = $sth->fetchrow_hashref();

	my $user = new PatchWork::User($row->{id});
	$user->username($row->{username});
	$user->name($row->{name});
	$user->email($row->{email});
	return $user;
}

###############################################################################
#
# Look up a user by username
#
###############################################################################
sub get_user_by_username($$)
{
	my ($obj, $username) = @_;
	return $obj->_lookup_user('users.username = ?', ($username));
}

###############################################################################
#
# Add a user
#
###############################################################################
sub add_user($$$)
{
	my ($obj, $user, $password) = @_;

	my $person = $obj->lookup_person($user->email(), $user->name());

	$user->id($person->id());

	$obj->_query('INSERT INTO users' .
		     " (id, username, passwd) VALUES (".
		     $user->id() . "," .
		     "'" . $user->username() . "'," .
		     "'" . md5_hex($password) . "')"
		     );

#	$obj->_query('INSERT INTO users'.
#			' (id, username, passwd) VALUES (?,?,?)',
#			$user->id(), $user->username(), md5_hex($password));

	$obj->commit();
}

###############################################################################
#
# Change a user's password
#
###############################################################################
sub change_password($$$)
{
	my ($obj, $user, $password) = @_;

	$obj->_query('UPDATE users SET passwd=? WHERE username=?',
		     md5_hex($password),
		     $user->username());

	$obj->commit();
}

###############################################################################
#
# Load the list of known upstream trees from the database
#
###############################################################################
sub _load_tree_list($)
{
	my ($obj) = @_;

	unless ($obj->{trees}) {
		my $sth = $obj->_query('SELECT id, name, url FROM trees');
		my %trees;

		while (my $row = $sth->fetchrow_hashref()) {
			my $tree = new PatchWork::Tree($row->{id});
			$tree->name($row->{name});
			$tree->url($row->{url});
			$trees{$row->{id}} = $tree;
		}

		$obj->{trees} = \%trees;
	}
}

###############################################################################
#
# Get a list of known upstream trees to which the patch may have been submitted
#
###############################################################################
sub trees(@)
{
	my $obj = shift;

	$obj->_load_tree_list() unless $obj->{trees};

	if (@_) {
		my $id = shift;
		return $obj->{trees}->{$id};
	}

	return $obj->{trees};
}

###############################################################################
#
# Find a list of trees associated with a patch
#
###############################################################################
sub find_patch_trees($$)
{
	my ($obj, $patch) = @_;

	my $patchfile = $patch->patchfile();

	my $sth = $obj->_query('SELECT tree'.
			       ' FROM patch_trees'.
			       ' WHERE patchfile=?',
			       $patchfile->id());

	my %trees = ();

	while (my $row = $sth->fetchrow_hashref()) {
		$trees{$row->{tree}} = $obj->trees($row->{tree});
	}

	$patchfile->trees(\%trees);
}

###############################################################################
#
# Update the list of trees associated with a patch
# - caller must commit
#
###############################################################################
sub update_patch_trees($$$)
{
	my ($obj, $patch, $trees) = @_;

	my $patchfile = $patch->patchfile();

	foreach my $tree (keys %{$obj->trees()}) {
		my $newstate = exists $trees->{$tree};
		my $oldstate = exists $patchfile->trees()->{$tree};

		if ($newstate && !$oldstate) {
			$obj->_query('INSERT INTO patch_trees'.
				     ' (patchfile, tree) '.
				     ' VALUES (' .$patchfile->id().', '.$tree.')');
		}
		elsif (!$newstate && $oldstate) {
			$obj->_query('DELETE FROM patch_trees'.
				     ' WHERE patchfile=? AND tree=?',
				     $patchfile->id(), $tree);
		}
	}
}

###############################################################################
#
# Format a date
#
###############################################################################
sub _date_format($$)
{
	my ($obj, $date) = @_;
	my ($sec, $min, $hour, $mday, $mon, $year, undef, undef, undef) =
	    localtime($date);

	return sprintf("%4d-%02d-%02d %02d:%02d:%02d",
		       $year + 1900, $mon + 1, $mday,
		       $hour, $min, $sec);
}

###############################################################################
#
# Format a project database ID or substitute "NULL"
#
###############################################################################
sub _project_format($$)
{
	my ($obj, $project) = @_;

	return "NULL" unless $project;
	return $project->id();
}

###############################################################################
#
# Format a bug report database ID or substitute "NULL"
#
###############################################################################
sub _bug_format($$)
{
	my ($obj, $bug) = @_;

	return "NULL" unless $bug;
	return $bug->id();
}

1;
