# Patchwork - automated patch tracking system
# Copyright (C) 2005 Jeremy Kerr <jk@ozlabs.org>
# - Modified by David Howells <dhowells@redhat.com>
#
# This file is part of the Patchwork package.
#
# Patchwork is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Patchwork is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Patchwork; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package PatchWork::PatchParser;

use strict;
use PatchWork::Person;
use PatchWork::Submission;
use PatchWork::PatchFile;
use PatchWork::Patch;
use PatchWork::Comment;
use PatchWork::SiteConfig;
use MIME::Parser;
use Date::Parse;

# internal variables
#   config: PatchWork::SiteConfig
#   tmp          tmp should *only* be defined while we're in parseMIMEStructure
#      ->header  the top-level mime entity's header
#      ->patch   used while we're parsing - the patch is placed here
#      ->comment comments placed where while we're parsing and have no patch
#   patch	patch that has been extracted
#   comments	comments that have been extracted
#   projectname	assumed project name

sub new($$)
{
	my ($cls, $config) = @_;
	my $obj = {};
	bless($obj, $cls);
	$obj->{config} = $config;
	$obj->{submission} = ();
	$obj->{submitter} = ();
	$obj->{projectnames} = ();
	$obj->{patchfiles} = ();
	$obj->{comments} = ();
	$obj->{refs} = ();
	$obj->{bug} = ();
	$obj->{npatches} = 0;
	return $obj;
}

###############################################################################
#
# Read and parse an email message from a file
#
###############################################################################
sub parse_MIME_file($$$)
{
	my ($obj, $file, $options) = @_;

	my $mimeparser = new MIME::Parser();
	$mimeparser->output_to_core(1);

	my $entity = $mimeparser->parse_open($file);

	return $obj->parse_MIME_structure($entity, $options);
}

###############################################################################
#
# Read and parse an email message from a stream
#
###############################################################################
sub parse_MIME_stream($$$)
{
	my ($obj, $stream, $options) = @_;

	my $mimeparser = new MIME::Parser();
	$mimeparser->output_to_core(1);

	my $entity = $mimeparser->parse($stream);

	return $obj->parse_MIME_structure($entity, $options);
}

###############################################################################
#
# Parse an email message, which may be MIME encoded
# - options:
# 	UseCurTime:	use the current time rather that that in the email
# 	IgnoreListID:	don't abort the parse if the listID doesn't match
#
###############################################################################
sub parse_MIME_structure($$$)
{
	my ($obj, $entity, $options) = @_;

	unless (ref($entity) eq 'MIME::Entity') {
		warn 'Argument to parse_MIME_structure isn\'t a MIME::Entity';
		return undef;
	}

	# reset from previous parse
	$obj->{submitter} = ();
	$obj->{projectnames} = ();
	$obj->{patchfiles} = ();
	$obj->{comments} = ();
	$obj->{refs} = ();
	$obj->{bug} = ();

	my $header = $entity->head();

	# skip emails not from the specified list
	unless ($options->{IgnoreListID}) {
		my $listid = $header->get('List-ID') or
			$header->get('X-Mailing-List') or
			return undef;
		$listid = $obj->_clean_header($listid);

		my $regex = '<'.$obj->{config}->list_id().'>';
		return undef unless $listid =~ m/$regex/;
	}

	# parse each entity and its subparts
	my @entities = ($entity);
	while (my $e = pop(@entities)) {
		$obj->_parse_entity($e);
		push(@entities, $e->parts());
	}

	my @patchfiles = @{$obj->{patchfiles}} if $obj->{patchfiles};
	my @comments = @{$obj->{comments}} if $obj->{comments};

	return undef unless (@patchfiles || @comments);

	# grab general header info for patches and comments
	$obj->{submitter} = $obj->_parse_email_addr($obj->_clean_header($header->get('From')));

	my $date = $options->{UseCurTime} ? time() : str2time($header->get('Date'));

	# declare a submission descriptor if we've got at least one patch
	my $submission = undef;

	if (@patchfiles) {
		my $subject = $header->get('Subject');

		$submission = new PatchWork::Submission();
		$submission->subject($obj->_parse_subject($subject));
		$submission->msgid($obj->_clean_header($header->get('Message-Id')));
		$submission->date($date);
		$submission->npatches($obj->{npatches});

		# set the attributes on the patch files we've obtained
		$_->submission($submission) foreach (@patchfiles);

		$obj->{submission} = $submission;

		$obj->_extract_bugreport($header->get('Subject'));
		$obj->_extract_bugreport($_->content()) foreach (@comments);
	}

	# set the attributes on the comments we've obtained
	foreach my $comment (@comments) {
		$comment->msgid($obj->_clean_header($header->get('Message-Id')));
		$comment->date($date);
		$comment->submission($submission);
		next unless $comment->content() =~ m/\S+/;
	}

	# we need references for comments without a patch
	unless (@patchfiles) {
		my $ref = $header->get('References');
		return undef unless $ref;
		my @refs = split(/\s+/, $ref);
		$obj->{refs} = \@refs;
	}

	return 1;
}

###############################################################################
#
# Parse a address for simple email address and name components
#
###############################################################################
sub _parse_email_addr($$)
{
	my ($obj, $addr) = @_;
	my ($email, $name) = ();

	if ($addr =~ m/"?(.*?)"?\s*<([^>]+)>/) {
		$email = $2;
		$name = $1;
	}
	elsif ($addr =~ m/"?(.*?)"?\s*\(([^\)]+)\)/) {
		$email = $1;
		$name = $2;
	}
	else {
		$email = $addr;
		$name = $addr;
	}

	return [$email, $name];
}

###############################################################################
#
# Parse a part of the email, as defined by MIME boundaries
#
###############################################################################
sub _parse_entity($$)
{
	my ($obj, $entity) = @_;

	if ($entity->effective_type() eq 'text/x-patch') {
		my $patchfile = new PatchWork::PatchFile();
		$patchfile->patchix(++$obj->{npatches});
		$patchfile->filename($entity->head()->recommended_filename());
		$patchfile->content($entity->bodyhandle()->as_string());
		push @{$obj->{patchfiles}}, $patchfile;
	}

	if ($entity->effective_type() eq 'text/plain') {
		$obj->_parse_text($entity->bodyhandle()->as_string());
	}

}

###############################################################################
#
# Parse a text part of the email that may have either comments and/or embedded
# patches
#
###############################################################################
sub _parse_text($$)
{
	my ($obj, $text) = @_;

	# $state specifies the last line we saw.
	my $state = 0;
	# 0: text
	# 1: suspected patch header (diff, ====, Index:)
	# 2: patch header line 1 (---)
	# 3: patch header line 2 (+++)
	# 4: patch hunk header line (@@ line)
	# 5: patch hunk content
	#
	# valid transitions:
	#  0 -> 1 (diff, ===, Index:)
	#  0 -> 2 (---)
	#  1 -> 2 (---)
	#  2 -> 3 (+++)
	#  3 -> 4 (@@ line)
	#  4 -> 5 (patch content)
	#  5 -> 1 (run out of lines from @@-specifed count)
	#
	# Suspected patch header is stored into $buf, and appended to
	# $patchbuf if we find a following hunk. Otherwise, append to
	# $commentbuf after parsing.

	my $patchbuf = '';
	my $commentbuf = '';
	my $buf = '';
	my @c = (0, 0);

	my $hunk = 0;

	foreach (split(/\n/, $text)) {
		$_ .= "\n";
		# handy debug mode:
		#$_ .= " ($state $hunk " . $c[0] . ":" . $c[1] . ")\n";

	      reconsider:

		if ($state == 0) {
			# looking for the start of a patch
			if (m/^diff / || m/^====+ / || m/^Index: /) {
				$state = 1;
				$buf .= $_;
			} elsif (m/^--- /) {
				$state = 2;
				$buf .= $_;
			} else {
				$commentbuf .= $_;
			}
		}
		elsif ($state == 1) {
			# may be viewing patch prologue
			$buf .= $_;
			if (m/^--- /) {
				$state = 2;
			}
		}
		elsif ($state == 2) {
			# got "---" line; looking for "+++" line
			if (m/^\+{3} /) {
				$state = 3;
				$buf .= $_;
			} elsif ($hunk) {
				$state = 1;
				$buf .= $_;
			} else {
				$state = 0;
				$commentbuf .= $buf.$_;
				$buf = '';
			}
		}
		elsif ($state == 3) {
			# looking for '@@' line
			if (m/^\@\@ -\d+(?:,(\d+))? \+\d+(?:,(\d+))? \@\@/) {
				@c = ($1 || 1, $2 || 1);
				$state = 4;
				$patchbuf .= $buf.$_;
				$buf = '';
			} elsif ($hunk) {
				$state = 1;
				goto reconsider;
			} else {
				$state = 0;
				$commentbuf .= $buf.$_;
				$buf = '';
			}
		}
		elsif ($state == 4 || $state == 5) {
			if (m/^\-/) {
				$c[0]--;
			} elsif (m/^\+/) {
				$c[1]--;
			} else {
				$c[0]--;
				$c[1]--;
			}

			$patchbuf .= $_;
			if ($c[0] <= 0 && $c[1] <= 0) {
				$state = 3;
				$hunk++;
			} else {
				$state = 5;
			}
		}
		else {
			warn "Unknown state (state=$state, line='$_')";
			$state = 0;
		}
	}

	if ($patchbuf) {
		my $patchfile = new PatchWork::PatchFile();
		$patchfile->patchix(++$obj->{npatches});
		$patchfile->content($patchbuf);
		push @{$obj->{patchfiles}}, $patchfile;
	}

	# any remaining mid-patch data wasn't followed by a complete patch
	$commentbuf .= $buf;

	if ($commentbuf) {
		my $re = $obj->{config}->list_footer();
		$commentbuf =~ s/\s*$re\s*//g if $re;
		$commentbuf =~ s/^\s//sg;
		my $comment = new PatchWork::Comment();
		$comment->content($commentbuf);
		push @{$obj->{comments}}, $comment;
	}
}

###############################################################################
#
# Scan the comments for the first bug report reference
#
###############################################################################
sub _extract_bugreport($$)
{
	my ($obj, $content) = @_;

	return $obj->{bug} if $obj->{bug};

	while ($content =~ m/\b((rhbz|bz|bug|bugzilla)|(rhfz|fz|feature|featurezilla)|(rhit|it|issue|issue\s+tracker))\s*([\#]|\s|)(\d+)\b/gi) {
		if (!$obj->{bug}) {
			my $bug = new PatchWork::Bug;
			$bug->set("BZ", $6) if $2;
			$bug->set("FZ", $6) if $3;
			$bug->set("IT", $6) if $4;
			$obj->{bug} = $bug;
		}
	}
}

###############################################################################
#
# Attempt to determine the name of the project to which the patch belongs from
# the contents of either the message subject or the comments in the patch
# message body
#
###############################################################################
sub _determine_projects($$)
{
	my ($obj, $str) = @_;

	while ($str =~ m/\bFC(\s*|-)(\d+)\b/gi) {
		push @{$obj->{projectnames}}, "FC-$2";
	}

	if ($str =~ m/\bRHEL\s*[4]\b/i ||
	    $str =~ m/\bRHEL-4\b/i ||
	    $str =~ m/\bNAHANT\b/i
	    ) {
		push @{$obj->{projectnames}}, "RHEL-4";
	}

	if ($str =~ m/\bRHEL\s*3\b/i ||
	    $str =~ m/\bRHEL-3\b/i ||
	    $str =~ m/\bTAROON\b/i
	    ) {
		push @{$obj->{projectnames}}, "RHEL-3";
	}

	if ($str =~ m/\bRHEL\s*21\b/i ||
	    $str =~ m/\bRHEL\s*2[.]1\b/i ||
	    $str =~ m/\bRHEL-21\b/i ||
	    $str =~ m/\bRHEL-2[.]1\b/i
	    ) {
		push @{$obj->{projectnames}}, "RHEL-2.1";
	}
}

###############################################################################
#
# Parse the subject
#
###############################################################################
sub _parse_subject($$)
{
	my ($obj, $str) = @_;
	chomp $str;

	# remove the tag inserted into the subject by the list expander
	if (my $re = $obj->{config}->subject_regex()) {
		$str =~ s/\s*$re\s*//gi;
	}

	$obj->_determine_projects($str);

	$str =~ s/\s*\[PATCH\]\s*//gi;
	return $str;
}

###############################################################################
#
# Clean up a header line value
#
###############################################################################
sub _clean_header($$)
{
	(my $obj, $_) = @_;
	return '' unless $_;
	chomp;
	s/^\s*//g;
	s/\s*$//g;
	return $_;
}

sub submitter($)
{
	my ($obj) = @_;
	return @{$obj->{submitter}};
}

sub submission(@)
{
	my ($obj) = shift;
	if (@_) { $obj->{submission} = shift }
	return $obj->{submission};
}

sub patchfiles($)
{
	my ($obj) = @_;
	return $obj->{patchfiles} ? @{$obj->{patchfiles}} : ();
}

sub projectnames($)
{
	my ($obj) = @_;
	return $obj->{projectnames} ? @{$obj->{projectnames}} : ();
}

sub comments($)
{
	my ($obj) = @_;
	return $obj->{comments} ? @{$obj->{comments}} : ();
}

sub refs($)
{
	my ($obj) = @_;
	return $obj->{refs} ? @{$obj->{refs}} : ();
}

sub bug($)
{
	my ($obj) = @_;
	return $obj->{bug};
}

1;
