# Patchwork - automated patch tracking system
# Copyright (C) 2005 Jeremy Kerr <jk@ozlabs.org>
# - Modified by David Howells <dhowells@redhat.com>
#
# This file is part of the Patchwork package.
#
# Patchwork is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Patchwork is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Patchwork; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package PatchWork::Web;

use strict;
use PatchWork::Web::Auth;

my $BZURL = 'https://bugzilla.redhat.com/bugzilla';

sub new($$)
{
	my ($cls, $config) = @_;
	my $obj = {};
	bless($obj, $cls);
	$obj->{config} = $config;
	$obj->{db} = new PatchWork::PatchDatabase($config);
	$obj->{auth} = new PatchWork::Web::Auth($obj);
	$obj->_parse_params();
	return $obj;
}

sub authenticate($$)
{
	my ($obj, $mandatory) = @_;
	unless ($obj->{auth}->authenticate($obj->params(), $mandatory)) {
		print $obj->{auth}->authcontent();
		exit;
	}
}

sub header($$)
{
	my ($obj, $title) = @_;
	my $file = $obj->{config}->web_include_dir().'/header.html';
	my $content = $obj->{auth} ? $obj->{auth}->header() : '';
	$content .= "Content-Type: text/html\n\n";
	open(FH, '<', $file) or die "Couldn't open $file: $!";
	{
		local $/;
		$content .= <FH>;
	}
	close(FH);
	my %subst = (title => $title, auth => $obj->_auth_string());
	$content =~ s/#$_#/$subst{$_}/eg for keys %subst;
	return $content;
}

sub footer($)
{
	my ($obj) = @_;
	my $file = $obj->{config}->web_include_dir().'/footer.html';
	my $content;
	open(FH, '<', $file) or die "Couldn't open $file: $!";
	{
		local $/;
		$content = <FH>;
	}
	close(FH);
	return $content;
}

sub _clean_param(@)
{
	(my $obj, $_) = @_;
	tr/+/ /;
	s/%([a-f0-9]{2})/chr(hex($1))/ieg;
	s/^\s+//;
	s/\s$//;
	return $_;
}

sub _parse_params($)
{
	my ($obj) = @_;

	my $buf = undef;

	if (!$ENV{REQUEST_METHOD} || $ENV{REQUEST_METHOD} eq 'GET') {
		$buf = $ENV{QUERY_STRING};
		
	} elsif ($ENV{REQUEST_METHOD} eq 'POST') {
		read(STDIN, $buf, $ENV{CONTENT_LENGTH})
			or die "Couldn't read post data: $!";

	} else {
		return;
	}

	my $values;
	if ($buf) {
		foreach my $pair (split(/&/, $buf)) {
			my ($name, $value) = split(/=/, $pair);
			map { $obj->_clean_param($_) } ($name, $value);
			$values->{$name} = $value;
		}
	}

	$buf = $ENV{HTTP_COOKIE};
	if ($buf) {
		foreach my $pair (split(/;/, $buf)) {
			my ($name, $value) = split(/=/, $pair);
			map { $obj->_clean_param($_) } ($name, $value);
			$values->{$name} = $value;
		}
	}
	
	$obj->{params} = $values;
	return $obj->{params};
}

sub params(@)
{
	my $obj = shift;
	return $obj->{params} unless @_;
	my $name = shift;
	return $obj->{params}->{$name};
}

sub param(@)
{
	my ($obj) = shift;
	return $obj->params(@_);
}

sub format_date($$)
{
	my ($obj, $date) = @_;
	my @time = localtime($date);
	return (sprintf("%04d-%02d-%02d",
		$time[5]+1900, $time[4]+1, $time[3]));
}

sub format_time($$)
{
	my ($obj, $date) = @_;
	my @time = localtime($date);
	return (sprintf("%04d-%02d-%02d %02d:%02d:%02d",
		$time[5]+1900, $time[4]+1, $time[3],
		$time[2], $time[1], $time[0]));
}

sub format_person($$)
{
	my ($obj, $person) = @_;
	my $email = $obj->obfuscate_email($person->email());
	my $linktext = $person->name();
	$linktext = $email unless $linktext;
	return sprintf '<a href="mailto:%s">%s</a>', $email, $linktext;
}

sub format_project($$)
{
	my ($obj, $project) = @_;

	return $project ? $project->pretty() : "&nbsp;";
}

sub format_bug($$)
{
	my ($obj, $bug) = @_;

	return "&nbsp;" unless $bug;

	my $ref = $bug->ref();
	if ($bug->type() eq 'BZ') {
		return '<a href="' . $BZURL . '/show_bug.cgi?id=' . $ref .
		    '">BZ ' . $ref . '</a>';
	}
	elsif ($bug->type() eq 'FZ') {
		return '<a href="' . $BZURL . '/show_bug.cgi?id=' . $ref .
		    '">FZ ' . $ref . '</a>';
	}
	elsif ($bug->type() eq 'IT') {
		return '<a href="' .
		    'https://enterprise.redhat.com/portal/?module=issues&action=view&tid=' .
		    $ref . '">IT ' . $ref . '</a>';
	}
	else {
		return $bug->type() . ' ' . $ref;
	}
}

sub obfuscate_email($$)
{
	(my $obj, $_) = @_;
	s/\@/ at /g, s/\./ dot /g;
	return $_;
}

sub auth(@)
{
	my ($obj) = shift;
	if (@_) { $obj->{auth} = shift }
	return $obj->{auth};
}

sub _auth_string($)
{
	my ($obj) = @_;
	my $str = '';
	return $str unless $obj->{auth};

	if ($obj->{auth}->authenticated()) {
		$str = sprintf("Logged in as <b>%s</b>",
			$obj->auth->user->name());
		$str .= " (<a href=\"changepassword.pl\">change password</a>)"
	} else {
		$str = sprintf('<a href="%s">Log in</a>',
			$obj->auth->login_link()) if
			$obj->auth->login_link();
	}
	return $str;
}

sub self($)
{
	my ($obj) = @_;
	my ($self) = ($0 =~ m#/([^/]+)$#);
	return $self;
}

sub db($)
{
	my ($obj) = @_;
	return $obj->{db};
}

sub url_encode(@)
{
	(my $obj, $_) = @_;
	# from RFC 2396, Section 2.3
	my $reserved = qr([^-a-z0-9_\.!~\*'\(\)]);
	s/($reserved)/'%'.hex(ord($1))/ieg;
	return $_;
}

sub escape_html($)
{
	(my $obj, $_) = @_;
	s/&/&amp;/g;
	s/</&lt;/g;
	s/>/&gt;/g;
	s/"/&quot;/g;
	return $_;
}

1;
