# Patchwork - automated patch tracking system
# Copyright (C) 2005 Jeremy Kerr <jk@ozlabs.org>
#
# This file is part of the Patchwork package.
#
# Patchwork is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Patchwork is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Patchwork; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package PatchWork::Web::Auth;

use strict;

sub new($$)
{
	my ($cls, $web) = @_;
	my $obj = {};
	$cls .= '::'.$web->{config}->web_auth_type();
	bless($obj, $cls);
	$obj->{web} = $web;
	$obj->init();
	return $obj;
}

sub init($)
{
	my ($obj) = @_;
	$obj->{user} = $obj->{authenticated} = undef;
}

sub authenticated($)
{
	my ($obj) = @_;
	return $obj->{authenticated};
}

sub user($)
{
	my ($obj) = @_;
	return $obj->{user};
}

# process the authentication data for the CGI.
# parameter:
# 	$vars - get or post vars for this page
# 	$mandatory - the page requires authentication
# returns false if we need to stop the normal page processing here. (ie, the
# authentication is mandatory and we can't authenticate). In this case, the
# caller should print the return value of authcontent() and exit.
sub authenticate($$$)
{
	return 1;
}

# header() returns header data to be inserted into the stream of the output
# page, regardless of whether the request is authenticated.
sub header($)
{
	return '';
}

# This function will print content required to start authentication. If the
# authenticate method fails, the web object should print the resulting string
# and exit.
sub authcontent($)
{
	return '';
}

# This function allows the Web module to provide a link to log in with
sub login_link($)
{
	return 'https://host198.cambridge.redhat.com';
}

# Our first authentication module: HTTPSBasic. If the connection is over HTTPS,
# we will require authentication, using the HTTP Basic method. If we're using
# plain HTTP, we can continue unauthenticated.
{
	package PatchWork::Web::Auth::HTTPSBasic;
	@PatchWork::Web::Auth::HTTPSBasic::ISA = ('PatchWork::Web::Auth');

	sub authenticate($$$)
	{
		my ($obj, $params, $mandatory) = @_;
		$obj->{mandatory} = $mandatory || defined $ENV{HTTPS}; 
		return 1 unless $ENV{HTTPS};
		return undef unless $ENV{REMOTE_USER};

		$obj->{user} = $obj->{web}->db->get_user_by_username(
			$ENV{REMOTE_USER});

		$obj->{authenticated} = defined($obj->{user});
		return !($obj->{mandatory} && !$obj->{authenticated});
	}

	sub authcontent($)
	{
		my ($obj) = @_;
		return qq(Status: 401 Unauthorized\n).
			qq(WWW-Authenticate: Basic realm=").
				$obj->{web}->{config}->name().
				qq( Patch Tracking System"\n).
			qq(Content-type: text/html\n\n).
			qq(<html><head><title>401 Unauthorized</title></head>).
			qq(<body><h1>401 Unauthorized</h1><p>You are not ).
			qq(authorized to view this page</p></body></html>\n);
	}

}


# next authentication module: Session. Uses the CGI::Session module to create
# our own session, containing the UID.
{
	package PatchWork::Web::Auth::Session;
	@PatchWork::Web::Auth::Session::ISA = ('PatchWork::Web::Auth');

	use CGI::Session;
	use File::Spec;

	my $timeout = 20 * 60;
	my $sessiondir = File::Spec->tmpdir."/patchwork";

	sub authenticate($$$)
	{
		my ($obj, $params, $mandatory) = @_;
		$obj->{mandatory} = $mandatory;
		my $session;

		$obj->{is_login} = $obj->{web}->params('action') && 
			$obj->{web}->params('action') eq 'login';
		if ($obj->{is_login}) {
			$session = $obj->_session_login() or return undef;

		} elsif (my $sid = $obj->{web}->params('sid')) {
			$session = new CGI::Session(undef, $sid,
				{Directory => $sessiondir});
		}

		$obj->{user} = $obj->{web}->{db}->get_user_by_id(
				$session->param('uid')) if $session;
		$obj->{authenticated} = defined $obj->{user};
		$obj->{session} = $session;

		return undef if $obj->{is_login};
			
		return !($obj->{mandatory} && !$obj->{authenticated});
	}

	sub _session_login($)
	{
		my ($obj) = @_;

		$obj->{user} = $obj->{web}->{db}->get_user_by_auth(
			$obj->{web}->param("username"),
			$obj->{web}->param("password"));

		unless ($obj->{user}) {
			$obj->{errstr} = 'Invalid username or password'
				if $obj->{web}->param("username");
			return undef;
		}

		my $session = new CGI::Session(undef, undef,
				{Directory => $sessiondir});
		$session->param('uid', $obj->{user}->id());
		$session->expire($timeout);

		return $session;
	}

	sub authcontent($)
	{
		my ($obj) = @_;
		if ($obj->{is_login} && $obj->authenticated()) {
			return sprintf("Set-Cookie: sid=%s\n".
				"Location: index.pl\n\n",
				$obj->{session}->id());
		}
		my $content = $obj->{web}->header('Authentication Required');
		if ($obj->{errstr}) {
			$content .= "<p>".$obj->{errstr}."</p>";
		}
		$content .= qq(<form method="post">).
		qq(<input type="hidden" name="action" value="login">).
		qq(<table class="login">).
		qq(<tr><th>Username:</th>).
		qq(<td><input name="username"></td></tr>).
		qq(<tr><th>Password:</th>).
		qq(<td><input type="password" name="password"></td></tr>).
		qq(<tr><td colspan="2" align="center">).
		qq(<input type="submit" value="Login"></td></tr></table></form);
		$content .= $obj->{web}->footer();
		return $content;
	}

	sub header($)
	{
		my ($obj) = @_;
		return '' unless ($obj->{session});
		return sprintf("Set-Cookie: sid=%s;\n", $obj->{session}->id());
	}

	sub login_link($)
	{
		my ($obj) = @_;
		return $obj->{web}->self()."?action=login";
	}

}


1;
