#!/usr/bin/perl -wT
#
# Patchwork - automated patch tracking system
# Copyright (C) 2005 Jeremy Kerr <jk@ozlabs.org>
# - Modified by David Howells <dhowells@redhat.com>
#
# This file is part of the Patchwork package.
#
# Patchwork is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Patchwork is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Patchwork; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use strict;
use lib '../lib';

use PatchWork::SiteConfig;
use PatchWork::PatchDatabase;
use PatchWork::Web;
use PatchWork::Web::Auth;

my $config;
my $w;
my %states;
my %restrictions;

my $archived;
my $offset;
my $limit;
my $order;
my $submitter;
my $project;
my $filter;
my $bugfilter;

my $xlimit;
my $xoffset;
my $xsubmitter;
my $xproject;
my $xorder;

my %people;
my %projects;
my @patches;

###############################################################################
#
# Generate an href to index.pl
#
###############################################################################
sub gen_index_href($$)
{
	my ($change, $to) = @_;
	my $sep = '?';
	my $str = "index.pl"; # $w->self();

	foreach my $key (keys(%restrictions)) {
		if ($key eq $change) {
			if ($to) {
				$str .= $sep . $change . '=' . $to;
				$sep = '&';
			}
			$change = "";
		}
		else {
			if ($restrictions{$key}) {
				$str .= $sep . $key . '='. $restrictions{$key};
				$sep = '&';
			}
		}
	}
	
	$str .= $sep . $change . '=' . $to if ($change && $to);

	return $str;
}

###############################################################################
#
# Display the view restrictions form
#
###############################################################################
sub display_view_restrictions_form()
{
	my $i;
	my $j;
	my $highlight;

	printf("<form action=\"%s\" method=\"get\">\n", gen_index_href("limit", undef));
	printf("  <input type=\"hidden\" name=\"order\" value=\"%s\"/>\n", $order) if $order;
	printf("  <table cellspacing=\"1\" cellpadding=\"1\" border=\"0\">\n");

	# view window start
	printf("    <tr>\n");
	printf("      <td align=right>First patch:</td>\n");
	printf("      <td><input name=\"offset\" %s/></td>\n",
	       $offset ? "value=\"". $offset . "\"" : "");
	printf("    </tr>\n");

	# view window size
	printf("    <tr>\n");
	printf("      <td align=right>View limit:</td>\n");
	printf("      <td><input name=\"limit\" %s/></td>\n",
	       $limit ? "value=\"". $limit . "\"" : "");
	printf("    </tr>\n");
	printf("    <tr>\n");

	# archived patch set
	printf("      <td align=right>Patch set:</td>\n");
	printf("      <td><select name=\"archived\" id=\"archived\">\n");
	printf("        <option value=\"0\" %s>Unarchived%s</option>\n",
	       $archived ? "" : "selected",
	       $archived ? "" : "*");
	printf("        <option value=\"1\" %s>Archived%s</option>\n",
	       $archived ? "selected" : "",
	       $archived ? "*" : "");
	printf("      </select></td>\n");
	printf("    </tr>\n");

	# project restriction list
	printf("      <td align=right>Project:</td>\n");
	printf("      <td><select name=\"project\" id=\"project\">\n");
	printf("        <option value=\"All\" %s>All%s</option>\n",
	       $xproject eq 'All' ? "selected" : "",
	       $xproject eq 'All' ? "*" : "");

	foreach $i (sort keys %projects) {
		$j = $projects{$i}->id();
		$highlight = ($xproject eq $j);
		printf("        <option value=\"%u\" %s>%s%s</option>\n",
		       $j,
		       $highlight ? "selected" : "",
		       $i,
		       $highlight ? "*" : "");
	}
	printf("      </select></td>\n");
	printf("    </tr>\n");

	# submitter restriction list (in name order)
	printf("    <tr>\n");
	printf("      <td align=right>Submitter:</td>\n");
	printf("      <td><select name=\"submitter\" id=\"submitter\">\n");
	printf("        <option value=\"All\" %s>All%s</option>\n",
	       $xsubmitter eq 'All' ? "selected" : "",
	       $xsubmitter eq 'All' ? "*" : "");

	my %xp = map { ($people{$_}->name() => $_) } keys %people;
	for $i (sort keys %xp) {
		$j = $people{$xp{$i}}->id();
		printf("        <option value=\"%u\" %s>%s%s</option>\n",
		       $j,
		       $xsubmitter eq $j ? "selected" : "",
		       $i,
		       $xsubmitter eq $j ? "*" : "");
	}
	printf("      </select></td>\n");
	printf("    </tr>\n");

	# name filter
	printf("    <tr>\n");
	printf("      <td align=right>Patch subject filter:</td>\n");
	printf("      <td><input name=\"filter\" %s/></td>\n",
	       $filter ? "value=\"". $filter . "\"" : "");
	printf("    </tr>\n");

	# bug filter
	print("    <tr>\n");
	print("      <td align=right>Bug filter:</td>\n");
	print("      <td>");
	printf("        <input name=\"bugfilter\" %s/>\n",
	       $bugfilter ? "value=\"". $bugfilter . "\"" : "");
	print("Specify a list of types, ref IDs, \"set\" and \"unset\", using % as a wildcard");
	print("      </td>\n");
	print("    </tr>\n");

	# submission button
	print("    <tr>\n");
	print("      <td></td>\n");
	print("      <td>\n");
	print("        <input id=\"button\" type=\"submit\" value=\"Restrict view\"/>");
	print("      </td>\n");
	print("    </tr>\n");

	print("  </table>\n");
	print("</form>\n");
}

###############################################################################
#
# Display the patch table header line
#
###############################################################################
sub display_patch_table_header()
{
	print("<tr>\n");

	foreach my $col (qw(subject patch date submitter project ref state)) {
		my $str;
		print "  <th>";

		if ($col eq $xorder) {
			printf('<span class="colactive">%s</span>', ucfirst($col));
		} else {
			printf('<a class="colinactive" href="%s">%s</a>',
			       gen_index_href("order", $col),
			       ucfirst($col));
		}

		print "</th>\n";
	}

	print "  <th>&nbsp;</th></tr>\n";
}

###############################################################################
#
# Display a link to a patch-specific view
#
###############################################################################
sub patch_subject_link($)
{
	my ($patch) = @_;
	return sprintf('<a href="patch.pl?id=%d">%s</a>',
		       $patch->id(), $patch->patchfile()->submission()->subject());
}

###############################################################################
#
# Display a row describing a single patch
#
###############################################################################
my $rowcount = 0;
sub display_patch_row($)
{
	my ($patch) = @_;

	my $patchfile = $patch->patchfile();
	my $submission = $patchfile->submission();

	printf(" <tr class=\"%s\">\n", $rowcount++ % 2 ? 'odd' : 'even');

	my $col = "  <td>%s</td>\n";

	printf($col, patch_subject_link($patch));

	if ($submission->npatches() > 1) {
		printf($col, $patchfile->patchix() . '/' . $submission->npatches());
	}
	else {
		printf($col, "&nbsp;");
	}

	printf($col, $w->format_date($submission->date()));
	printf($col, $w->format_person($submission->submitter()));
	printf($col, $w->format_project($patch->project()));
	printf($col, $w->format_bug($patch->bug()));
	printf($col, $states{$patch->state()});

	printf('  <td>'.
	       '    <a href="patchcontent.pl?id=%d">' .
	       '      <img width="16" height="16" src="images/download.png" '.
	       '           border="0" alt="download" title="download"/>'.
	       '    </a>'.
	       "  </td>\n",
	       $patch->id());
	print " </tr>\n";
}

###############################################################################
#
# Gather information about the request
#
###############################################################################
$config = new PatchWork::SiteConfig() or die "Couldn't load configuration";
$w = new PatchWork::Web($config);
%states = $w->db()->get_states();

$w->authenticate(undef);

$archived	= $w->param('archived');
$offset		= $w->param('offset');
$limit		= $w->param('limit');
$order		= $w->param('order');
$project	= $w->param('project');
$submitter	= $w->param('submitter');
$filter		= $w->param('filter');
$bugfilter	= $w->param('bugfilter');

$archived = undef	if ($archived	&& $archived !~ m/^1$/);
$offset = undef		if ($offset	&& $offset !~ m/^[1-9]\d*$/);
$limit = undef		if ($limit	&& $limit !~ m/^[1-9]\d*$/);
$submitter = undef	if ($submitter	&& $submitter !~ m/^[1-9]\d*$/);
$project = undef	if ($project	&& $project !~ m/^[1-9]\d*$/);

$xlimit = $limit ? $limit : $offset ? 100 : undef;
$xoffset = $offset ? $offset : $xlimit ? 1 : undef;
$xorder = $order || 'state';
$xsubmitter = $submitter ? $submitter : "All";
$xproject = $project ? $project : "All";

$restrictions{archived}		= $archived;
$restrictions{offset}		= $xoffset;
$restrictions{limit}		= $xlimit;
$restrictions{order}		= $order;
$restrictions{project}		= $project;
$restrictions{submitter}	= $submitter;
$restrictions{filter}		= $filter;
$restrictions{bugfilter}	= $bugfilter;

###############################################################################
#
# Construct the web page
#
###############################################################################
print $w->header($config->name() . " Patches");

%people = $w->db()->get_people_list();
%projects = $w->db()->get_project_list();

display_view_restrictions_form();

@patches = $w->db()->get_patches(\%restrictions);

print "<table class=\"patchlist\">\n";

display_patch_table_header();
display_patch_row($_) foreach (@patches);

print "</table>";

print $w->footer();
