#!/usr/bin/perl -wT
#
# Patchwork - automated patch tracking system
# Copyright (C) 2005 Jeremy Kerr <jk@ozlabs.org>
# - Modified by David Howells <dhowells@redhat.com>
#
# This file is part of the Patchwork package.
#
# Patchwork is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Patchwork is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Patchwork; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use strict;
use lib '../lib';

use PatchWork::SiteConfig;
use PatchWork::PatchDatabase;
use PatchWork::Submission;
use PatchWork::PatchFile;
use PatchWork::Patch;
use PatchWork::Web;
use PatchWork::Web::Auth;

my $config;
my $w;
my %states;
my $submission;
my $patchfile;
my $patch;

###############################################################################
#
# Display the archived state
#
###############################################################################
sub display_archived($)
{
	my ($patch) = @_;
	my $archived = $patch->archived();

	return $archived ? "Archived" : "Unarchived"
	    unless $w->auth->authenticated();

	my $str .= "<select name=\"archived\" id=\"archived\">\n";
	$str .= sprintf("<option value=\"0\" %s>Unarchived%s</option>\n",
			$archived ? "" : "selected",
			$archived ? "" : "*");
	$str .= sprintf("<option value=\"1\" %s>Archived%s</option>\n",
			$archived ? "selected" : "",
			$archived ? "*" : "");
	$str .= "</select>\n";
	return $str;
}

###############################################################################
#
# Display the patch's state
#
###############################################################################
sub display_state($)
{
	my ($patch) = @_;
	return $states{$patch->state()} unless $w->auth->authenticated();

	my $str = '<select name="state">';
	my $fmt = '<option value="%d"%s>%s</option>';

	foreach my $i (sort keys %states) {
		my $highlight = ($i eq $patch->state());

		$str .= sprintf("<option value=\"%u\" %s>%s%s</option>\n",
				$i,
				$highlight ? "selected" : "",
				$states{$i},
				$highlight ? "*" : "");
	}

	$str .= '</select>';
	return $str;
}

###############################################################################
#
# Display the patch's tree information
#
###############################################################################
sub display_trees($)
{
	my ($patch) = @_;

	my $patchfile = $patch->patchfile();

	my $str = '';

	my %alltrees = %{$w->db->trees()};
	my %pftrees = %{$patchfile->trees()};

	if ($w->auth->authenticated()) {
		foreach my $treeid (sort keys %alltrees) {
			my $tree = $alltrees{$treeid};

			$str .= sprintf('<input type="checkbox" name="tree:%d"%s/>'.
					'%s<br>',
					$tree->id(),
					exists $pftrees{$treeid} ?
					' checked=1' : '',
					$tree->name());
		}
	}
	else {
		return 'none' unless $patchfile->trees();
		$str = join(',',
			    map { $pftrees{$_}->name() } sort keys(%pftrees));
	}

	return $str;
}

###############################################################################
#
# Display the project
#
###############################################################################
sub display_project($)
{
	my ($patch) = @_;

	return $w->format_project($patch->project())
	    unless $w->auth->authenticated();

	my %projects = $w->db()->get_project_list();

	my $str = "<select name=\"project\" id=\"project\">";

	my $xproject = $patch->project() ? $patch->project()->id() : 'Unset';

	$str .= sprintf("<option value=\"Unset\" %s>Unset%s</option>\n",
			$xproject eq 'Unset' ? "selected" : "",
			$xproject eq 'Unset' ? "*" : "");

	foreach my $i (sort keys %projects) {
		my $j = $projects{$i}->id();
		my $highlight = ($xproject eq $j);

		$str .= sprintf("<option value=\"%u\" %s>%s%s</option>\n",
				$j,
				$highlight ? "selected" : "",
				$i,
				$highlight ? "*" : "");
	}

	$str .= "</select>";

	return $str;
}

###############################################################################
#
# Display the bug report
#
###############################################################################
sub display_bug($)
{
	my ($patch) = @_;

	return $w->format_bug($patch->bug())
	    unless $w->auth->authenticated();

	my $bug = $patch->bug();
	my $str = "";

#	my $xtype = $bug ? $bug->type() : "Unset";

	# bug report type selection
#	my $str = "<select name=\"bugtype\" id=\"bugtype\">";
#
#	foreach my $type ("Unset", "BZ", "FZ", "IT") {
#		$str .= sprintf("<option value=\"%s\" %s>%s%s</option>\n",
#				$type,
#				$type eq $xtype ? "selected" : "",
#				$type,
#				$type eq $xtype ? "*" : "");
#	}
#
#	$str .= "</select>";

	$str .= sprintf("<input name=\"bugtype\" %s/>&nbsp;",
			defined $bug ? "value=\"" . $bug->type() . "\"" : "");

	$str .= sprintf("<input name=\"bugref\" %s/>&nbsp;",
			defined $bug ? "value=\"" . $bug->ref() . "\"" : "");

	$str .= $w->format_bug($patch->bug());

	return $str;
}

###############################################################################
#
# Display the file download point
#
###############################################################################
sub display_file_download($$)
{
	my ($patch, $patchfile) = @_;


	my $filename = $patchfile->filename();
	$filename = 'patchwork-' . $patchfile->id() . '.patch' unless $filename;

	return sprintf('<a href="patchcontent.pl?id=%d">%s</a>',
		       $patch->id(), $filename);
}

###############################################################################
#
# Display the patch's comments
#
###############################################################################
sub comment_html($)
{
	my ($comment) = @_;

	print '<div class="comment">';
	printf('<div class="meta">%s %s</div>',
	       $w->format_person($comment->submitter()),
	       $w->format_time($comment->date()));

	my $content = $w->escape_html($comment->content());
	$content =~ s!^(\s*&gt;.*)$!<span class="quote">$1</span>!mg;
	print '<pre class="content">' . $content . '</pre>';
	print '</div>';
}

###############################################################################
#
# Gather information about the request
#
###############################################################################
$config = new PatchWork::SiteConfig() or die "Couldn't load configuration";
$w = new PatchWork::Web($config);
%states = $w->db()->get_states();

$w->authenticate(undef);

###############################################################################
#
# Grab the patch
#
###############################################################################
$patch = $w->db->get_patch($w->params('id'));
unless ($patch) {
	print $w->header("Patch Tracking System");
	print ("Error: Patch not found\n");
	print $w->footer();
	exit;
}

$patchfile = $patch->patchfile();
$submission = $patchfile->submission();

$w->db->find_patch_trees($patch);

print $w->header("Patch: ".$w->escape_html($submission->subject()));
print(qq(<form action="setpatch.pl" method="post">\n));
printf(qq(<input type="hidden" name="id" value="%d"/>\n), $patch->id());
print(qq(<table class="patchmeta">\n));

my %values = (
	'Patch attachment'	=> $patchfile->patchix() . ' of ' . $submission->npatches(),
	'Submitter'		=> $w->format_person($submission->submitter()),
	'Date'			=> $w->format_time($submission->date()),
	'Message ID'		=> $w->escape_html($submission->msgid()),
	'Project'		=> display_project($patch),
	'Primary bug reference'	=> display_bug($patch),
	'Download'		=> display_file_download($patch, $patchfile),
	'State'			=> display_state($patch),
	'Upstream acceptance'	=> display_trees($patch),
	'Patch set'		=> display_archived($patch),
);

printf("<tr><th>%s:</th><td>%s</td></tr>\n", $_, $values{$_})
    for ('Submitter',
	 'Date',
	 'Message ID',
	 'Patch attachment',
	 'Download',
	 'Project',
	 'Primary bug reference',
	 'Patch set',
	 'State',
	 'Upstream acceptance');

if ($w->auth->authenticated()) {
	print "<tr><td></td><td>\n";
	print(qq(<input type="submit" name="action" value="Update"/>&nbsp;));
	print "</td></tr>\n";
}

print("</table>\n");

if ($w->auth->authenticated()) {
	print(qq(<input type="submit" name="action" value="Delete Patch" ).
		qq(onClick="return confirmDelete()"/>));
}

print("</form>\n");

###############################################################################
#
# display the comments
#
###############################################################################
print('<h2><a name="patch">Comments</a></h2>');

my @comments = $w->db()->get_comments($submission);
comment_html($_) for @comments;
@comments = undef;


###############################################################################
#
# display the patch
#
###############################################################################
print '<h2><a name="patch">Patch</a></h2>';
print '<div class="patch">';
print "<pre class=\"content\">";

my $newpatch = 1;
print "<div class=\"header\">";

foreach (split '\n', $patchfile->content()) {
	if (/^(---|\+\+\+) ([^\t]*)(.*)/) {
		print "<div class=\"header\">" if !$newpatch;
		print $1, " <font color=blue>", $w->escape_html($2), "</font>";
		print $w->escape_html($3);
		$newpatch = 1;
	}
	elsif (/^@@ (-\d+,\d+) (\+\d+,\d+) @@(.*)$/) {
		print "</div>" if $newpatch;
		print "<b>@@ <font color=orangered>$1</font> <font color=green>$2</font> @@</b>";
		print $w->escape_html($3) if $3;
		$newpatch = 0;
	}
	elsif (/^([-+ ])/) {
		print "<font color=orangered>" if $1 eq '-';
		print "<font color=green>" if $1 eq '+';
		print $w->escape_html($_);
		print "</font>" if $1 ne ' ';
	}
	elsif (/^(===*) (.*) (===*)$/) {
		print "<div class=\"header\">" if !$newpatch;
		print $1, " <font color=blue>", $w->escape_html($2), "</font> ", $3;
		$newpatch = 1;
	}
	else {
		print "<div class=\"header\">" if !$newpatch;
		print $w->escape_html($_);
		$newpatch = 1;
	}

	print "\n";
}

print "</div>" if $newpatch;
print "</pre>\n";
print '</div>';

print $w->footer();
